@@ -51,36 +51,38 @@
  * REGISTER is special because it may tie resources on the server side */
 #define TPSIP_CONNECTION_MINIMUM_KEEPALIVE_INTERVAL_REGISTER 50
 
-/* The value of TPSIP_NH_EXPIRED. This can be anything that is neither NULL
- * nor a media channel */
-NUA_HMAGIC_T * const _tpsip_nh_expired = (NUA_HMAGIC_T *)"";
-
-static sip_to_t *priv_sip_to_url_make (TpsipConnection *conn,
-                                       su_home_t *home,
-                                       TpHandle contact)
+static sip_to_t *
+priv_sip_to_url_make (TpsipConnection *conn,
+                      su_home_t *home,
+                      TpHandle contact)
 {
-  TpHandleRepoIface *contact_repo;
-  sip_to_t *to;
-  const char *address;
+  const url_t *url;
 
-  contact_repo = tp_base_connection_get_handles (
-      (TpBaseConnection *)conn, TP_HANDLE_TYPE_CONTACT);
+  url = tpsip_conn_get_contact_url (conn, contact);
+  return sip_to_create (home, (const url_string_t *) url);
+}
 
-  address = tp_handle_inspect (contact_repo, contact);
-  if (address == NULL)
+static sip_from_t *
+priv_sip_from_url_make (TpsipConnection *conn,
+                        su_home_t *home)
+{
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (conn);
+  sip_from_t *from;
+  gchar *alias = NULL;
+
+  from = sip_from_create (home, (const url_string_t *) priv->account_url);
+
+  if (from == NULL)
     return NULL;
 
-  /* TODO: set display name bound to the handle using qdata? */
+  g_object_get (conn, "alias", &alias, NULL);
+  if (alias != NULL)
+    from->a_display = su_strdup (home, alias);
 
-  to = sip_to_create (home, URL_STRING_MAKE(address));
-
-  if (to &&
-      url_sanitize(to->a_url) == 0) 
-    return to;
-
-  return NULL;
+  return from;
 }
 
+
 nua_handle_t *
 tpsip_conn_create_register_handle (TpsipConnection *conn,
                                  TpHandle contact)
@@ -105,25 +107,25 @@ tpsip_conn_create_register_handle (TpsipConnection *conn,
 
 nua_handle_t *
 tpsip_conn_create_request_handle (TpsipConnection *conn,
-                                TpHandle contact)
+                                  TpHandle contact)
 {
   TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (conn);
   nua_handle_t *result = NULL;
   su_home_t temphome[1] = { SU_HOME_INIT(temphome) };
+  sip_from_t *from;
   sip_to_t *to;
 
   g_assert (priv->sofia_home != NULL);
   g_assert (priv->sofia_nua != NULL);
 
   to = priv_sip_to_url_make (conn, temphome, contact);
+  from = priv_sip_from_url_make (conn, temphome);
 
-  /* TODO: Pass also SIPTAG_FROM updated from base->self_handle, to update the
-   * display name possibly set by the client */
-
-  if (to)
+  if (to != NULL && from != NULL)
     result = nua_handle (priv->sofia_nua, NULL,
                          NUTAG_URL(to->a_url),
                          SIPTAG_TO(to),
+                         SIPTAG_FROM(from),
                          TAG_END());
 
   su_home_deinit (temphome);
@@ -657,247 +659,259 @@ tpsip_conn_discover_stun_server (TpsipConnection *conn)
 }
 
 static gboolean
-priv_is_user_unreserved (gchar x)
+priv_is_host (const gchar* str)
 {
-    switch (x)
-      {
-        case '-':
-        case '_':
-        case '.':
-        case '!':
-        case '~':
-        case '*':
-        case '\'':
-        case '(':
-        case ')':
-        case '&':
-        case '=':
-        case '+':
-        case '$':
-        case ',':
-        case ':':
-        case '?':
-        case ';':
-        case '/':
-          return TRUE;
-        default:
-          return g_ascii_isalnum (x);
-      }
+  static GRegex *host_regex = NULL;
+
+#define DOMAIN "[a-z0-9]([-a-z0-9]*[a-z0-9])?"
+#define TLD "[a-z]([-a-z0-9]*[a-z0-9])?"
+
+  if (host_regex == NULL)
+    {
+      GError *error = NULL;
+
+      host_regex = g_regex_new ("^("
+            "("DOMAIN"\\.)*"TLD"\\.?|"      /* host name */
+            "[0-9]{1,3}(\\.[0-9]{1,3}){3}|" /* IPv4 address */
+            "\\[[0-9a-f:.]\\]"              /* IPv6 address, sloppily */
+          ")$",
+          G_REGEX_CASELESS | G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, &error);
+
+      if (error != NULL)
+        g_error ("failed to compile the host regex: %s", error->message);
+    }
+
+#undef DOMAIN
+#undef TLD
+
+  return g_regex_match (host_regex, str, 0, NULL);
 }
 
 static gboolean
-priv_is_host (gchar x)
+priv_is_tel_num (const gchar *str)
 {
-    switch (x)
-      {
-        case '.':
-        case '-':
-          return TRUE;
-        default:
-          return g_ascii_isalnum (x);
-      }
+  static GRegex *tel_num_regex = NULL;
+
+  if (tel_num_regex == NULL)
+    {
+      GError *error = NULL;
+
+      tel_num_regex = g_regex_new (
+          "^\\s*[\\+(]?\\s*[0-9][-.0-9()\\s]*$",
+          G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, &error);
+
+      if (error != NULL)
+        g_error ("failed to compile the telephone number regex: %s", error->message);
+    }
+
+  return g_regex_match (tel_num_regex, str, 0, NULL);
 }
 
+/* Strip the non-essential characters from a string regarded as
+ * a telephone number */
 static gchar *
-priv_user_encode (su_home_t *home, const gchar *string)
+priv_strip_tel_num (const gchar *fuzzy)
 {
-  const gchar *a;
-  gchar *b;
-  gchar *res = su_zalloc (home, strlen (string) * 3 + 1);
+  static GRegex *cruft_regex = NULL;
 
-  g_return_val_if_fail (res != NULL, NULL);
-
-  a = string;
-  b = res;
-  while (*a)
+  if (cruft_regex == NULL)
     {
-      if (priv_is_user_unreserved (*a))
+      GError *error = NULL;
+
+      cruft_regex = g_regex_new ("[^+0-9]+",
+          G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, &error);
+
+      if (error != NULL)
+        g_error ("failed to compile the non-essential telephone number cruft regex: %s", error->message);
+    }
+
+  return g_regex_replace_literal (cruft_regex, fuzzy, -1, 0, "", 0, NULL);
+}
+
+static const char *
+priv_lowercase_url_part (su_home_t *home, const char *src)
+{
+  size_t n = 0;
+  size_t i;
+  char *res;
+
+  for (i = 0; src[i]; i++)
+    {
+      if (g_ascii_isupper (src[i]))
         {
-          *b++ = *a++;
-        }
-      else
-        {
-          snprintf (b, 4, "%%%02x", (guint) *a);
-          ++a;
-          b += 3;
+          n = i + strlen (src + i);
+          break;
         }
     }
 
-  return res;
+  if (!src[i])
+    return src;
+
+  res = su_alloc (home, n + 1);
+  memcpy (res, src, i);
+  for (; i < n; i++)
+    res[i] = g_ascii_tolower (src[i]);
+  res[i] = '\0';
+
+  return (const char *) res;
 }
 
-/* unescape characters that don't need escaping */
-static gchar *
-priv_user_decode (su_home_t *home, const gchar *string)
-{
-    const gchar *a;
-    gchar *b;
-    gchar *res = su_zalloc (home, strlen (string) + 1);
-
-    g_return_val_if_fail (res != NULL, NULL);
-
-    a = string;
-    b = res;
-    while (*a)
-      {
-        if ((a[0] == '%') && g_ascii_isxdigit(a[1]) && g_ascii_isxdigit(a[2]))
-          {
-            gchar tmp[3] = { a[1], a[2], 0 };
-            gchar x = (gchar) (strtoul (tmp, NULL, 16));
-            if (priv_is_user_unreserved (x))
-              {
-                *b++ = x;
-                a += 3;
-                continue;
-              }
-          }
-        *b++ = *a++;
-      }
-
-    return res;
-}
-
-static gboolean
-priv_is_tel_num (const gchar *string)
-{
-  const gchar *pc;
-  gboolean has_digits = FALSE;
-
-  g_return_val_if_fail (string != NULL, FALSE);
-
-  /* skip the initial whitespace */
-  pc = string + strspn (string, " \t");
-
-  /* the leading '+' is acceptable */
-  if (*pc == '+')
-    ++pc;
-
-  /* only digits, delimiters and inline whitespace */
-  while (*pc)
-    {
-      if (g_ascii_isdigit (*pc))
-        has_digits = TRUE;
-      else
-        switch (*pc)
-          {
-          case ' ':
-          case '\t':
-          case '-':
-          case '.':
-          case '(':
-          case ')':
-            break;
-          default:
-            return FALSE;
-          }
-      ++pc;
-    }
-
-  return has_digits;
-}
-
-static gchar *
-priv_strip_whitespace (su_home_t *home, const gchar *string)
-{
-  const gchar *a;
-  gchar *b;
-  gchar *res = su_zalloc (home, strlen (string) + 1);
-
-  g_return_val_if_fail (res != NULL, NULL);
-
-  b = res;
-  for (a = string; *a; a++)
-    {
-      if (!g_ascii_isspace (*a))
-        *b++ = *a;
-    }
-  *b = '\0';
-
-  return res;
-}
+#define TPSIP_RESERVED_CHARS_ALLOWED_IN_USERNAME "!*'()&=+$,;?/"
 
 gchar *
-tpsip_conn_normalize_uri (TpsipConnection *conn,
+tpsip_handle_normalize (TpHandleRepoIface *repo,
                         const gchar *sipuri,
+                        gpointer context,
                         GError **error)
 {
+  TpsipConnection *conn = TPSIP_CONNECTION (context);
   TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (conn);
-  su_home_t home[1] = { SU_HOME_INIT (home) };
-  url_t *url = NULL;;
+  const url_t *base_url = priv->account_url;
+  su_home_t home[1] = { SU_HOME_INIT(home) };
+  url_t *url;
   gchar *retval = NULL;
-  char *c, *str;
+  char *c;
 
   url = url_make (home, sipuri);
 
-  /* we got username or phone number, local to our domain */
-  if ((url == NULL) ||
-      ((url->url_scheme == NULL) && (url->url_user == NULL)))
+  if (url == NULL ||
+      (url->url_scheme == NULL && url->url_user == NULL))
     {
-      if ((priv->account_url == NULL) || (priv->account_url->url_host == NULL))
+      /* we got username or phone number, local to our domain */
+      gchar *user;
+
+      if (base_url == NULL || base_url->url_host == NULL)
         {
-          g_debug ("local uri specified and we don't know local domain yet");
+          g_warning ("bare name given, but no account URL is set");
           goto error;
         }
 
       if (priv_is_tel_num (sipuri))
         {
+          user = priv_strip_tel_num (sipuri);
           url = url_format (home, "sip:%s@%s;user=phone",
-              priv_strip_whitespace (home, sipuri),
-              priv->account_url->url_host);
+              user, base_url->url_host);
         }
       else
         {
+          user = g_uri_escape_string (sipuri,
+              TPSIP_RESERVED_CHARS_ALLOWED_IN_USERNAME, FALSE);
           url = url_format (home, "sip:%s@%s",
-              priv_user_encode (home, sipuri),
-              priv->account_url->url_host);
+              user, base_url->url_host);
         }
+
+      g_free (user);
+
       if (!url) goto error;
     }
-  else
-    {
-      if ((url != NULL) && (url->url_user != NULL))
-        {
-          url->url_user = (char *) priv_user_decode (home, url->url_user);
-        }
-    }
 
-  if (url_sanitize (url)) goto error;
+  if (url_sanitize (url) != 0) goto error;
 
-  /* scheme and host should've been set by now */
-  if (!url->url_scheme || (url->url_scheme[0] == 0) ||
-      !url->url_host || (url->url_host[0] == 0))
-      goto error;
+  /* scheme should've been set by now */
+  if (url->url_scheme == NULL || (url->url_scheme[0] == 0))
+    goto error;
 
-  for (c = (char *) url->url_host; *c; c++)
-    {
-      /* check for illegal characters */
-      if (!priv_is_host (*c))
-          goto error;
+  /* convert the scheme to lowercase */
+  /* Note: we can't do it in place because url->url_scheme may point to
+   * a static string */
+  url->url_scheme = priv_lowercase_url_part (home, url->url_scheme);
 
-      /* convert host to lowercase */
-      *c = g_ascii_tolower (*c);
-    }
-  /* check that the hostname isn't empty */
-  if (c == url->url_host) goto error;
-
-  /* check that if we have '@', the username isn't empty */
+  /* Check that if we have '@', the username isn't empty.
+   * Note that we rely on Sofia-SIP to canonize the user name */
   if (url->url_user)
     {
       if (url->url_user[0] == 0) goto error;
     }
 
-  str = url_as_string (home, url);
-  if (NULL == str) goto error;
+  /* host should be set and valid */
+  if (url->url_host == NULL || !priv_is_host (url->url_host))
+      goto error;
 
-  retval = g_strdup (str);
+  /* convert host to lowercase */
+  for (c = (char *) url->url_host; *c; c++)
+    {
+      *c = g_ascii_tolower (*c);
+    }
+
+  retval = g_strdup (url_as_string (home, url));
 
 error:
-  if (NULL == retval)
-      g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
+  if (retval == NULL)
+      g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_HANDLE,
           "invalid SIP URI");
 
-  /* success */
   su_home_deinit (home);
   return retval;
 }
 
+static GQuark
+tpsip_handle_url_quark ()
+{
+  static GQuark quark = 0;
+
+  if (G_UNLIKELY (quark == 0))
+    quark = g_quark_from_static_string ("tpsip-handle-url");
+
+  return quark;
+}
+
+const url_t*
+tpsip_conn_get_contact_url (TpsipConnection *self,
+                            TpHandle handle)
+{
+  TpBaseConnection *base = (TpBaseConnection *) self;
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_handles;
+  GQuark url_quark;
+  url_t *url;
+  GError *error;
+
+  contact_handles = tp_base_connection_get_handles (base,
+      TP_HANDLE_TYPE_CONTACT);
+
+  if (!tp_handle_is_valid (contact_handles, handle, &error))
+    {
+      DEBUG("invalid handle %u: %s", handle, error->message);
+      g_error_free (error);
+      return NULL;
+    }
+
+  url_quark = tpsip_handle_url_quark ();
+
+  url = tp_handle_get_qdata (contact_handles, handle, url_quark);
+
+  if (url == NULL)
+    {
+      url = url_make (priv->sofia_home,
+          tp_handle_inspect (contact_handles, handle));
+
+      tp_handle_set_qdata (contact_handles, handle, url_quark, url, NULL);
+    }
+
+  return url;
+}
+
+TpHandle
+tpsip_handle_parse_from (TpHandleRepoIface *contact_repo,
+                         const sip_t *sip)
+{
+  TpHandle handle = 0;
+  gchar *url_str;
+
+  g_return_val_if_fail (sip != NULL, 0);
+
+  if (sip->sip_from)
+    {
+      su_home_t tmphome[1] = { SU_HOME_INIT(tmphome) };
+
+      url_str = url_as_string (tmphome, sip->sip_from->a_url);
+
+      handle = tp_handle_ensure (contact_repo, url_str, NULL, NULL);
+
+      /* TODO: set qdata for the display name */
+
+      su_home_deinit (tmphome);
+    }
+
+  return handle;
+}
