@@ -33,6 +33,7 @@
 #include <telepathy-glib/handle-repo-dynamic.h>
 #include <telepathy-glib/interfaces.h>
 #include <telepathy-glib/svc-connection.h>
+#include <telepathy-glib/svc-generic.h>
 
 #include <tpsip/event-target.h>
 
@@ -40,13 +41,14 @@
 #include "media-factory.h"
 #include "text-factory.h"
 
+#include "conn-aliasing.h"
+
 #include "sip-connection-enumtypes.h"
 #include "sip-connection-helpers.h"
 #include "sip-connection-private.h"
 #include "sip-connection-sofia.h"
 
 #include <sofia-sip/msg_header.h>
-#include <sofia-sip/sip_status.h>
 
 #define DEBUG_FLAG TPSIP_DEBUG_CONNECTION
 #include "debug.h"
@@ -57,6 +59,12 @@ static void conn_iface_init (gpointer, gpointer);
 G_DEFINE_TYPE_WITH_CODE(TpsipConnection, tpsip_connection,
     TP_TYPE_BASE_CONNECTION,
     G_IMPLEMENT_INTERFACE (TPSIP_TYPE_EVENT_TARGET, event_target_iface_init);
+    G_IMPLEMENT_INTERFACE(TP_TYPE_SVC_DBUS_PROPERTIES,
+        tp_dbus_properties_mixin_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CONNECTION_INTERFACE_CONTACTS,
+        tp_contacts_mixin_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CONNECTION_INTERFACE_ALIASING,
+        tpsip_conn_aliasing_iface_init);
     G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CONNECTION, conn_iface_init))
 
 #define ERROR_IF_NOT_CONNECTED_ASYNC(BASE, CONTEXT) \
@@ -76,6 +84,7 @@ enum
   PROP_ADDRESS = 1,      /**< public SIP address (SIP URI) */
   PROP_AUTH_USER,        /**< account username (if different from public address userinfo part) */
   PROP_PASSWORD,         /**< account password (for registration) */
+  PROP_ALIAS,               /* Display name for self */
 
   PROP_TRANSPORT,        /**< outbound transport */
   PROP_PROXY,            /**< outbound SIP proxy (SIP URI) */
@@ -123,82 +132,56 @@ priv_url_from_string_value (su_home_t *home, const GValue *value)
   return (url_str)? url_make (home, url_str) : NULL;
 }
 
-static TpHandle
-priv_handle_parse_from (const sip_t *sip,
-                        su_home_t *home,
-                        TpHandleRepoIface *contact_repo)
-{
-  TpHandle handle = 0;
-  gchar *url_str;
-
-  g_return_val_if_fail (sip != NULL, 0);
-
-  if (sip->sip_from)
-    {
-      url_str = url_as_string (home, sip->sip_from->a_url);
-
-      handle = tp_handle_ensure (contact_repo, url_str, NULL, NULL);
-
-      su_free (home, url_str);
-
-      /* TODO: set qdata for the display name */
-    }
-
-  return handle;
-}
-
-static gchar *normalize_sipuri (TpHandleRepoIface *repo, const gchar *sipuri,
-    gpointer context, GError **error);
-
 static void
 tpsip_create_handle_repos (TpBaseConnection *conn,
-                         TpHandleRepoIface *repos[NUM_TP_HANDLE_TYPES])
+                           TpHandleRepoIface *repos[NUM_TP_HANDLE_TYPES])
 {
   repos[TP_HANDLE_TYPE_CONTACT] =
       (TpHandleRepoIface *)g_object_new (TP_TYPE_DYNAMIC_HANDLE_REPO,
           "handle-type", TP_HANDLE_TYPE_CONTACT,
-          "normalize-function", normalize_sipuri,
+          "normalize-function", tpsip_handle_normalize,
           "default-normalize-context", conn,
           NULL);
 }
 
 static GPtrArray *
-tpsip_connection_create_channel_factories (TpBaseConnection *base)
+tpsip_connection_create_channel_managers (TpBaseConnection *conn)
 {
-  TpsipConnection *self = TPSIP_CONNECTION (base);
-  TpsipConnectionPrivate *priv;
-  GPtrArray *factories = g_ptr_array_sized_new (2);
+  TpsipConnection *self = TPSIP_CONNECTION (conn);
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+  GPtrArray *channel_managers = g_ptr_array_sized_new (2);
 
-  g_assert (TPSIP_IS_CONNECTION (self));
-  priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+  g_ptr_array_add (channel_managers,
+      g_object_new (TPSIP_TYPE_TEXT_FACTORY,
+        "connection", self, NULL));
 
-  priv->text_factory = (TpChannelFactoryIface *)g_object_new (
-      TPSIP_TYPE_TEXT_FACTORY, "connection", self, NULL);
-  g_ptr_array_add (factories, priv->text_factory);
+  priv->media_factory = g_object_new (TPSIP_TYPE_MEDIA_FACTORY,
+        "connection", self, NULL);
+  g_ptr_array_add (channel_managers, priv->media_factory);
 
-  priv->media_factory = (TpChannelFactoryIface *)g_object_new (
-      TPSIP_TYPE_MEDIA_FACTORY, "connection", self, NULL);
-  g_ptr_array_add (factories, priv->media_factory);
-
-  return factories;
+  return channel_managers;
 }
 
 static void
-tpsip_connection_init (TpsipConnection *obj)
+tpsip_connection_init (TpsipConnection *self)
 {
-  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (obj);
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+
   priv->sofia_home = su_home_new(sizeof (su_home_t));
-  priv->auth_table = g_hash_table_new_full (g_direct_hash,
-                                            g_direct_equal,
-                                            NULL /* (GDestroyNotify) nua_handle_unref */,
-                                            g_free);
+
+  tp_contacts_mixin_init ((GObject *) self,
+      G_STRUCT_OFFSET (TpsipConnection, contacts));
+
+  tp_base_connection_register_with_contacts_mixin ((TpBaseConnection *) self);
+
+  tpsip_conn_aliasing_init (self);
 }
 
 static void
 tpsip_connection_set_property (GObject      *object,
-                             guint         property_id,
-                             const GValue *value,
-                             GParamSpec   *pspec)
+                               guint         property_id,
+                               const GValue *value,
+                               GParamSpec   *pspec)
 {
   TpsipConnection *self = (TpsipConnection*) object;
   TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
@@ -219,6 +202,11 @@ tpsip_connection_set_property (GObject      *object,
     priv->password = g_value_dup_string (value);
     break;
   }
+  case PROP_ALIAS: {
+    g_free(priv->alias);
+    priv->alias = g_value_dup_string (value);
+    break;
+  }
   case PROP_TRANSPORT: {
     g_free(priv->transport);
     priv->transport = g_value_dup_string (value);
@@ -230,7 +218,7 @@ tpsip_connection_set_property (GObject      *object,
   }
   case PROP_REGISTRAR: {
     priv->registrar_url = priv_url_from_string_value (priv->sofia_home, value);
-    if (priv->sofia_nua) 
+    if (priv->sofia_nua)
       nua_set_params(priv->sofia_nua,
                      NUTAG_REGISTRAR(priv->registrar_url),
                      TAG_END());
@@ -322,14 +310,18 @@ tpsip_connection_get_property (GObject      *object,
     g_value_set_string (value, priv->auth_user);
     break;
   }
-  case PROP_TRANSPORT: {
-    g_value_set_string (value, priv->transport);
-    break;
-  }
   case PROP_PASSWORD: {
     g_value_set_string (value, priv->password);
     break;
   }
+  case PROP_ALIAS: {
+    g_value_set_string (value, priv->alias);
+    break;
+  }
+  case PROP_TRANSPORT: {
+    g_value_set_string (value, priv->transport);
+    break;
+  }
   case PROP_PROXY: {
     priv_value_set_url_as_string (value, priv->proxy_url);
     break;
@@ -406,6 +398,12 @@ static gboolean tpsip_connection_start_connecting (TpBaseConnection *base,
 static void
 tpsip_connection_class_init (TpsipConnectionClass *klass)
 {
+  static const gchar *interfaces_always_present[] = {
+      TP_IFACE_CONNECTION_INTERFACE_REQUESTS,
+      TP_IFACE_CONNECTION_INTERFACE_CONTACTS,
+      TP_IFACE_CONNECTION_INTERFACE_ALIASING,
+      NULL };
+
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   TpBaseConnectionClass *base_class =
     (TpBaseConnectionClass *)klass;
@@ -414,11 +412,13 @@ tpsip_connection_class_init (TpsipConnectionClass *klass)
   /* Implement pure-virtual methods */
   base_class->create_handle_repos = tpsip_create_handle_repos;
   base_class->get_unique_connection_name = tpsip_connection_unique_name;
-  base_class->create_channel_factories =
-    tpsip_connection_create_channel_factories;
+  base_class->create_channel_managers =
+      tpsip_connection_create_channel_managers;
+  base_class->create_channel_factories = NULL;
   base_class->disconnected = tpsip_connection_disconnected;
   base_class->start_connecting = tpsip_connection_start_connecting;
   base_class->shut_down = tpsip_connection_shut_down;
+  base_class->interfaces_always_present = interfaces_always_present;
 
   g_type_class_add_private (klass, sizeof (TpsipConnectionPrivate));
 
@@ -431,180 +431,132 @@ tpsip_connection_class_init (TpsipConnectionClass *klass)
 #define INST_PROP(x) \
   g_object_class_install_property (object_class,  x, param_spec)
 
-  param_spec = g_param_spec_pointer("sofia-root",
-                                    "Sofia root",
-                                    "Event root from Sofia-SIP stack",
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_pointer ("sofia-root", "Sofia root",
+      "Event root from Sofia-SIP stack",
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_SOFIA_ROOT);
 
-  param_spec = g_param_spec_string("address",
-                                   "TpsipConnection construction property",
-                                   "Public SIP address",
-                                   NULL,
-                                   G_PARAM_CONSTRUCT_ONLY |
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("address", "SIP address",
+      "SIP AoR URI",
+      NULL,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_ADDRESS);
 
-  param_spec = g_param_spec_string("auth-user",
-                                   "Auth username",
-                                   "Username to use when registering (if different "
-                                   "than userinfo part of public SIP address)",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("auth-user", "Auth username",
+      "Username to use when registering "
+      "(if different than userinfo part of public SIP address)",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_AUTH_USER);
 
-  param_spec = g_param_spec_string("password",
-                                   "SIP account password",
-                                   "Password for SIP registration",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("password", "SIP account password",
+      "Password for SIP registration",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_PASSWORD);
 
-  param_spec = g_param_spec_string("transport",
-                                   "Transport protocol",
-                                   "Preferred transport protocol [optional]",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("alias", "Alias",
+      "User's display name",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  INST_PROP(PROP_ALIAS);
+
+  param_spec = g_param_spec_string ("transport", "Transport protocol",
+      "Preferred transport protocol (auto, udp, tcp)",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_TRANSPORT);
 
-  param_spec = g_param_spec_string("proxy",
-                                   "Outbound proxy",
-                                   "SIP URI for outbound proxy (e.g. 'sip:sipproxy.myprovider.com') [optional]",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("proxy", "Outbound proxy",
+      "SIP URI for outbound proxy",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_PROXY);
 
-  param_spec = g_param_spec_string("registrar",
-                                   "Registrar",
-                                   "SIP URI for registrar (e.g. 'sip:sip.myprovider.com') [optional]",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("registrar", "Registrar",
+      "SIP URI for registrar",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_REGISTRAR);
 
-  param_spec = g_param_spec_boolean("loose-routing",
-                                    "Loose routing",
-                                    "Enable loose routing as per RFC 3261",
-                                    FALSE, /*default value*/
-                                    G_PARAM_CONSTRUCT |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_boolean ("loose-routing", "Loose routing",
+      "Enable loose routing as per RFC 3261",
+      FALSE, /*default value*/
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_LOOSE_ROUTING);
 
-  param_spec = g_param_spec_enum ("keepalive-mechanism",
-                                  "Keepalive mechanism",
-                                  "SIP registration keepalive mechanism",
-                                  tpsip_connection_keepalive_mechanism_get_type (),
-                                  TPSIP_CONNECTION_KEEPALIVE_AUTO, /*default value*/
-                                  G_PARAM_CONSTRUCT |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_enum ("keepalive-mechanism", "Keepalive mechanism",
+      "Keepalive mechanism for SIP registration",
+      tpsip_connection_keepalive_mechanism_get_type (),
+      TPSIP_CONNECTION_KEEPALIVE_AUTO, /*default value*/
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_KEEPALIVE_MECHANISM);
 
-  param_spec = g_param_spec_int("keepalive-interval", 
-				"Keepalive interval",
-				"Interval between keepalives in seconds (0 = internal default, -1 = let the stack decide)",
-				-1, G_MAXINT32, -1,
-                                G_PARAM_CONSTRUCT |
-                                G_PARAM_READWRITE |
-                                G_PARAM_STATIC_NAME |
-                                G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_int ("keepalive-interval", "Keepalive interval",
+      "Interval between keepalives in seconds "
+      "(0 = internal default, -1 = let the stack decide)",
+      -1, G_MAXINT32, -1,
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_KEEPALIVE_INTERVAL);
 
-  param_spec = g_param_spec_boolean("discover-binding",
-                                    "Discover public contact",
-                                    "Enable discovery of public IP address beyond NAT",
-                                    TRUE, /*default value*/
-                                    G_PARAM_CONSTRUCT |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_boolean ("discover-binding",
+      "Discover public contact",
+      "Enable discovery of public IP address beyond NAT",
+      TRUE, /*default value*/
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_DISCOVER_BINDING);
 
-  param_spec = g_param_spec_boolean("discover-stun", "Discover STUN server",
-                                    "Enable discovery of STUN server host name "
-                                    "using DNS SRV lookup",
-                                    TRUE, /*default value*/
-                                    G_PARAM_CONSTRUCT |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_boolean ("discover-stun", "Discover STUN server",
+      "Enable discovery of STUN server host name using DNS SRV lookup",
+      TRUE, /*default value*/
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_DISCOVER_STUN);
 
-  param_spec = g_param_spec_string("stun-server",
-                                   "STUN server address",
-                                   "STUN server address (FQDN or IP address, "
-                                   "e.g. 'stun.myprovider.com') [optional]",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("stun-server", "STUN server address",
+      "STUN server address (FQDN or IP address)",
+      NULL,
+      G_PARAM_READWRITE |  G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_STUN_SERVER);
 
-  param_spec = g_param_spec_uint ("stun-port",
-                                  "STUN port",
-                                  "STUN port.",
-                                  0, G_MAXUINT16,
-                                  TPSIP_DEFAULT_STUN_PORT, /*default value*/
-                                  G_PARAM_CONSTRUCT |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_uint ("stun-port", "STUN port",
+      "STUN server port",
+      0, G_MAXUINT16,
+      TPSIP_DEFAULT_STUN_PORT, /*default value*/
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_STUN_PORT);
 
-  param_spec = g_param_spec_string("local-ip-address",
-                                   "Local IP address",
-                                   "Local IP address to use [optional]",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("local-ip-address", "Local IP address",
+      "Local IP address to use",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_LOCAL_IP_ADDRESS);
 
-  param_spec = g_param_spec_uint ("local-port",
-                                  "Local port",
-                                  "Local port for SIP [optional]",
-                                  0, G_MAXUINT16, 0,
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_uint ("local-port", "Local port",
+      "Local port for SIP",
+      0, G_MAXUINT16,
+      0,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_LOCAL_PORT);
 
-  param_spec = g_param_spec_string("extra-auth-user",
-                                   "Extra auth username",
-                                   "Username to use for extra authentication challenges",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("extra-auth-user", "Extra auth username",
+      "Username to use for extra authentication challenges",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_EXTRA_AUTH_USER);
 
-  param_spec = g_param_spec_string("extra-auth-password",
-                                   "Extra auth password",
-                                   "Password to use for extra authentication challenges",
-                                   NULL,
-                                   G_PARAM_READWRITE |
-                                   G_PARAM_STATIC_NAME |
-                                   G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_string ("extra-auth-password", "Extra auth password",
+      "Password to use for extra authentication challenges",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_EXTRA_AUTH_PASSWORD);
 
 #undef INST_PROP
+
+  tp_dbus_properties_mixin_class_init (object_class,
+      G_STRUCT_OFFSET (TpsipConnectionClass, properties_class));
+
+  tp_contacts_mixin_class_init (object_class,
+      G_STRUCT_OFFSET (TpsipConnectionClass, contacts_class));
 }
 
 static gboolean
@@ -701,10 +653,10 @@ priv_handle_auth (TpsipConnection* self,
   g_assert (realm != NULL);
   if (user && method) {
     if (realm[0] == '"')
-      auth = g_strdup_printf ("%s:%s:%s:%s", 
+      auth = g_strdup_printf ("%s:%s:%s:%s",
                               method, realm, user, password);
     else
-      auth = g_strdup_printf ("%s:\"%s\":%s:%s", 
+      auth = g_strdup_printf ("%s:\"%s\":%s:%s",
                               method, realm, user, password);
 
     DEBUG("%s authenticating user='%s' realm=%s",
@@ -795,189 +747,6 @@ tpsip_connection_nua_r_register_cb (TpsipConnection     *self,
   return TRUE;
 }
 
-static gboolean
-tpsip_connection_nua_i_invite_cb (TpsipConnection   *self,
-                                  const TpsipNuaEvent  *ev,
-                                  tagi_t             tags[],
-                                  gpointer           foo)
-{
-  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
-  TpsipMediaChannel *channel;
-  TpHandleRepoIface *contact_repo;
-  TpHandle handle;
-  GError *error = NULL;
-
-  /* figure out a handle for the identity */
-
-  contact_repo = tp_base_connection_get_handles ((TpBaseConnection *)self,
-                                                 TP_HANDLE_TYPE_CONTACT);
-
-  handle = priv_handle_parse_from (ev->sip, priv->sofia_home, contact_repo);
-  if (!handle)
-    {
-      g_message ("incoming INVITE with invalid sender information");
-      nua_respond (ev->nua_handle, 400, "Invalid From address", TAG_END());
-      return TRUE;
-    }
-
-  DEBUG("Got incoming invite from <%s>", 
-        tp_handle_inspect (contact_repo, handle));
-
-  channel = tpsip_media_factory_new_channel (
-                TPSIP_MEDIA_FACTORY (priv->media_factory),
-                NULL,
-                TP_HANDLE_TYPE_NONE,
-                0,
-                &error);
-  if (channel)
-    {
-      tpsip_media_channel_receive_invite (channel, ev->nua_handle, handle);
-    }
-  else
-    {
-      g_warning ("creation of SIP media channel failed: %s", error->message);
-      g_error_free (error);
-      nua_respond (ev->nua_handle, SIP_500_INTERNAL_SERVER_ERROR, TAG_END());
-    }
-
-  tp_handle_unref (contact_repo, handle);
-
-  return TRUE;
-}
-
-static gboolean
-tpsip_connection_nua_i_message_cb (TpsipConnection   *self,
-                                   const TpsipNuaEvent  *ev,
-                                   tagi_t             tags[],
-                                   gpointer           foo)
-{
-  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
-  TpsipTextChannel *channel;
-  TpHandleRepoIface *contact_repo;
-  TpHandle handle;
-  const sip_t *sip = ev->sip;
-  char *text = NULL;
-
-  /* Block anything else except text/plain messages (like isComposings) */
-  if (sip->sip_content_type
-      && (g_ascii_strcasecmp ("text/plain", sip->sip_content_type->c_type)))
-    {
-      nua_respond (ev->nua_handle,
-                   SIP_415_UNSUPPORTED_MEDIA,
-                   SIPTAG_ACCEPT_STR("text/plain"),
-                   NUTAG_WITH_THIS(ev->nua),
-                   TAG_END());
-      goto end;
-    }
-
-  /* If there is some text, assure it's in UTF-8 encoding */
-  if (sip->sip_payload && sip->sip_payload->pl_len > 0)
-    {
-      const char *charset = NULL;
-      if (sip->sip_content_type && sip->sip_content_type->c_params != 0)
-        {
-          charset = msg_params_find (sip->sip_content_type->c_params, "charset=");
-        }
-
-      /* Default charset is UTF-8, we only need to convert if it's a different one */
-      if (charset && g_ascii_strcasecmp (charset, "UTF-8"))
-        {
-          GError *error;
-          gsize in_len, out_len;
-          text = g_convert (sip->sip_payload->pl_data, sip->sip_payload->pl_len,
-              "UTF-8", charset, &in_len, &out_len, &error);
-
-          if (text == NULL)
-            {
-              gint status;
-              gchar *message;
-
-              status = (error->code == G_CONVERT_ERROR_ILLEGAL_SEQUENCE)
-                       ? 400 : 500;
-              message = g_strdup_printf ("Character set conversion failed"
-                                                " for the message body: %s",
-                                         error->message);
-              nua_respond (ev->nua_handle,
-                           status, message,
-                           NUTAG_WITH_THIS(ev->nua),
-                           TAG_END());
-
-              g_free (message);
-              g_error_free (error);
-              goto end;
-            }
-          if (in_len != sip->sip_payload->pl_len)
-            {
-              nua_respond (ev->nua_handle,
-                           400, "Incomplete character sequence at the "
-                                "end of the message body",
-                           NUTAG_WITH_THIS(ev->nua),
-                           TAG_END());
-              goto end;
-            }
-        }
-      else
-        {
-          if (!g_utf8_validate (sip->sip_payload->pl_data,
-                                sip->sip_payload->pl_len,
-                                NULL))
-            {
-              nua_respond (ev->nua_handle,
-                           400, "Invalid characters in the message body",
-                           NUTAG_WITH_THIS(ev->nua),
-                           TAG_END());
-              goto end;
-            }
-          text = g_strndup (sip->sip_payload->pl_data, sip->sip_payload->pl_len);
-        }
-    }
-  else
-    {
-      text = g_strdup ("");
-    }
-
-  contact_repo = tp_base_connection_get_handles (
-      (TpBaseConnection *)self, TP_HANDLE_TYPE_CONTACT);
-
-  handle = priv_handle_parse_from (sip, priv->sofia_home, contact_repo);
-
-  if (handle)
-    {
-      DEBUG("Got incoming message from <%s>", 
-            tp_handle_inspect (contact_repo, handle));
-
-      channel = tpsip_text_factory_lookup_channel (priv->text_factory, handle);
-
-      if (!channel)
-        {
-          channel = tpsip_text_factory_new_channel (priv->text_factory, handle,
-              NULL);
-          g_assert (channel != NULL);
-        }
-
-      tpsip_text_channel_receive (channel, handle, text);
-
-      tp_handle_unref (contact_repo, handle);
-
-      nua_respond (ev->nua_handle,
-                   SIP_200_OK,
-                   NUTAG_WITH_THIS(ev->nua),
-                   TAG_END());
-    }
-  else
-    {
-      nua_respond (ev->nua_handle,
-                   400, "Invalid From address",
-                   NUTAG_WITH_THIS(ev->nua),
-                   TAG_END());
-    }
-
-end:
-  g_free (text);
-
-  return TRUE;
-}
-
 static void
 tpsip_connection_shut_down (TpBaseConnection *base)
 {
@@ -1015,19 +784,16 @@ tpsip_connection_dispose (GObject *object)
 
   DEBUG("disposing of TpsipConnection %p", self);
 
-  /* these are borrowed refs, the real ones are owned by the superclass */
-  priv->media_factory = NULL;
-  priv->text_factory = NULL;
-
-  /* may theoretically involve NUA handle unrefs */
-  g_hash_table_destroy (priv->auth_table);
-
   /* the base class is responsible for unreffing the self handle when we
    * disconnect */
   g_assert (base->status == TP_CONNECTION_STATUS_DISCONNECTED
       || base->status == TP_INTERNAL_CONNECTION_STATUS_NEW);
   g_assert (base->self_handle == 0);
 
+  /* the base class owns channel factories/managers,
+   * here we just nullify the references */
+  priv->media_factory = NULL;
+
   if (G_OBJECT_CLASS (tpsip_connection_parent_class)->dispose)
     G_OBJECT_CLASS (tpsip_connection_parent_class)->dispose (object);
 }
@@ -1054,6 +820,7 @@ tpsip_connection_finalize (GObject *obj)
   g_free (priv->address);
   g_free (priv->auth_user);
   g_free (priv->password);
+  g_free (priv->alias);
   g_free (priv->transport);
   g_free (priv->stun_host);
   g_free (priv->local_ip_address);
@@ -1062,12 +829,14 @@ tpsip_connection_finalize (GObject *obj)
 
   g_free (priv->registrar_realm);
 
+  tp_contacts_mixin_finalize (obj);
+
   G_OBJECT_CLASS (tpsip_connection_parent_class)->finalize (obj);
 }
 
 static gboolean
 tpsip_connection_start_connecting (TpBaseConnection *base,
-                                 GError **error)
+                                   GError **error)
 {
   TpsipConnection *self = TPSIP_CONNECTION (base);
   TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
@@ -1081,10 +850,6 @@ tpsip_connection_start_connecting (TpBaseConnection *base,
   g_assert (priv->sofia_root != NULL);
   g_return_val_if_fail (priv->address != NULL, FALSE);
 
-  /* FIXME: we should defer setting the self handle until we've found out from
-   * the stack what handle we actually got, at which point we set it; and
-   * not tell Telepathy that connection has succeeded until we've done so
-   */
   contact_repo = tp_base_connection_get_handles (base, TP_HANDLE_TYPE_CONTACT);
   base->self_handle = tp_handle_ensure (contact_repo, priv->address,
       NULL, error);
@@ -1097,7 +862,7 @@ tpsip_connection_start_connecting (TpBaseConnection *base,
 
   DEBUG("self_handle = %d, sip_address = %s", base->self_handle, sip_address);
 
-  priv->account_url = url_make (priv->sofia_home, sip_address);
+  priv->account_url = tpsip_conn_get_contact_url (self, base->self_handle);
   if (priv->account_url == NULL)
     {
       g_set_error (error, TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
@@ -1149,14 +914,6 @@ tpsip_connection_start_connecting (TpBaseConnection *base,
    * at registration time */
   nua_get_params (priv->sofia_nua, TAG_ANY(), TAG_NULL());
 
-  g_signal_connect (self,
-                    "nua-event::nua_i_invite",
-                    G_CALLBACK (tpsip_connection_nua_i_invite_cb),
-                    NULL);
-  g_signal_connect (self,
-                    "nua-event::nua_i_message",
-                    G_CALLBACK (tpsip_connection_nua_i_message_cb),
-                    NULL);
   g_signal_connect (self,
                     "nua-event::nua_r_register",
                     G_CALLBACK (tpsip_connection_nua_r_register_cb),
@@ -1204,17 +961,6 @@ tpsip_connection_disconnected (TpBaseConnection *base)
     }
 }
 
-static gchar *
-normalize_sipuri (TpHandleRepoIface *repo,
-                  const gchar *sipuri,
-                  gpointer context,
-                  GError **error)
-{
-    TpsipConnection *conn = TPSIP_CONNECTION (context);
-
-    return tpsip_conn_normalize_uri (conn, sipuri, error);
-}
-
 static void
 conn_iface_init(gpointer g_iface, gpointer iface_data)
 {
