@@ -1,6 +1,6 @@
 /*
  * sip-text-channel.c - Source for TpsipTextChannel
- * Copyright (C) 2005-2007 Collabora Ltd.
+ * Copyright (C) 2005-2008 Collabora Ltd.
  * Copyright (C) 2005-2008 Nokia Corporation
  *   @author Martti Mela <first.surname@nokia.com>
  *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
@@ -23,27 +23,27 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <dbus/dbus-glib.h>
+#include "sip-text-channel.h"
+
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 
+#include <dbus/dbus-glib.h>
 #include <telepathy-glib/channel-iface.h>
 #include <telepathy-glib/dbus.h>
 #include <telepathy-glib/errors.h>
 #include <telepathy-glib/gtypes.h>
 #include <telepathy-glib/interfaces.h>
 #include <telepathy-glib/svc-channel.h>
+#include <telepathy-glib/svc-generic.h>
 
 #include <tpsip/event-target.h>
 
-#include "sip-text-channel.h"
-
 #include "sip-connection.h"
 #include "sip-connection-helpers.h"
 
-#include <sofia-sip/sip.h>
-#include <sofia-sip/sip_header.h>
+#include <sofia-sip/sip_status.h>
 
 #define DEBUG_FLAG TPSIP_DEBUG_IM
 #include "debug.h"
@@ -54,16 +54,27 @@ tpsip_text_channel_nua_r_message_cb (TpsipTextChannel *self,
                                      tagi_t            tags[],
                                      gpointer          foo);
 
-static void event_target_iface_init (gpointer, gpointer);
 static void channel_iface_init (gpointer, gpointer);
 static void text_iface_init (gpointer, gpointer);
+static void destroyable_iface_init (gpointer, gpointer);
 
 G_DEFINE_TYPE_WITH_CODE (TpsipTextChannel, tpsip_text_channel, G_TYPE_OBJECT,
-    G_IMPLEMENT_INTERFACE (TPSIP_TYPE_EVENT_TARGET, event_target_iface_init);
+    G_IMPLEMENT_INTERFACE (TPSIP_TYPE_EVENT_TARGET, NULL);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_DBUS_PROPERTIES,
+      tp_dbus_properties_mixin_iface_init);
     G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL, channel_iface_init);
     G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL_TYPE_TEXT, text_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL_INTERFACE_DESTROYABLE,
+      destroyable_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_EXPORTABLE_CHANNEL, NULL);
     G_IMPLEMENT_INTERFACE (TP_TYPE_CHANNEL_IFACE, NULL));
 
+static const char *tpsip_text_channel_interfaces[] = {
+    TP_IFACE_CHANNEL_INTERFACE_DESTROYABLE,
+    NULL
+};
+
+
 /* properties */
 enum
 {
@@ -72,7 +83,13 @@ enum
   PROP_CHANNEL_TYPE,
   PROP_HANDLE_TYPE,
   PROP_HANDLE,
-  /* PROP_CREATOR, */
+  PROP_TARGET_ID,
+  PROP_INITIATOR_HANDLE,
+  PROP_INITIATOR_ID,
+  PROP_REQUESTED,
+  PROP_INTERFACES,
+  PROP_CHANNEL_DESTROYED,
+  PROP_CHANNEL_PROPERTIES,
   LAST_PROPERTY
 };
 
@@ -84,15 +101,14 @@ typedef struct _TpsipTextPendingMessage TpsipTextPendingMessage;
 struct _TpsipTextPendingMessage
 {
   guint id;
-  nua_handle_t *nh;
-  
-  
   time_t timestamp;
   TpHandle sender;
-  
   TpChannelTextMessageType type;
-  
+  guint flags;
   gchar *text;
+
+  nua_handle_t *nh;
+  nua_saved_event_t saved_event[1];
 };
 
 typedef struct _TpsipTextChannelPrivate TpsipTextChannelPrivate;
@@ -102,11 +118,12 @@ struct _TpsipTextChannelPrivate
   TpsipConnection *conn;
   gchar *object_path;
   TpHandle handle;
+  TpHandle initiator;
 
   guint recv_id;
   guint sent_id;
   GQueue  *pending_messages;
-  GQueue  *messages_to_be_acknowledged;
+  GQueue  *sending_messages;
 
   gboolean closed;
 
@@ -114,26 +131,35 @@ struct _TpsipTextChannelPrivate
 };
 
 
-#define _tpsip_text_pending_new() \
-	(g_slice_new(TpsipTextPendingMessage))
 #define _tpsip_text_pending_new0() \
 	(g_slice_new0(TpsipTextPendingMessage))
 
 #define TPSIP_TEXT_CHANNEL_GET_PRIVATE(o)     (G_TYPE_INSTANCE_GET_PRIVATE ((o), TPSIP_TYPE_TEXT_CHANNEL, TpsipTextChannelPrivate))
 
-static void _tpsip_text_pending_free(TpsipTextPendingMessage *msg)
+static void tpsip_text_pending_free (TpsipTextPendingMessage *msg,
+                                     TpHandleRepoIface *contact_handles)
 {
-  if (msg->text)
-    g_free (msg->text);
+  if (msg->sender)
+    tp_handle_unref (contact_handles, msg->sender);
 
-  nua_handle_unref (msg->nh);
+  g_free (msg->text);
+
+  if (msg->saved_event[0])
+    nua_destroy_event (msg->saved_event);
+
+  if (msg->nh)
+    nua_handle_unref (msg->nh);
 
   g_slice_free (TpsipTextPendingMessage, msg);
 }
 
-static void _tpsip_text_pending_free_walk(gpointer data, gpointer user_data)
+static inline void
+tpsip_text_pending_respond (TpsipTextPendingMessage *msg)
 {
-  _tpsip_text_pending_free ((TpsipTextPendingMessage *) data);
+  nua_respond (msg->nh,
+               SIP_200_OK,
+               NUTAG_WITH_SAVED(msg->saved_event),
+               TAG_END());
 }
 
 static void
@@ -144,25 +170,31 @@ tpsip_text_channel_init (TpsipTextChannel *obj)
   DEBUG("enter");
 
   priv->pending_messages = g_queue_new ();
-  priv->messages_to_be_acknowledged = g_queue_new ();
+  priv->sending_messages = g_queue_new ();
 }
 
-static
-GObject *tpsip_text_channel_constructor(GType type,
-				      guint n_props,
-				      GObjectConstructParam *props)
+static void
+tpsip_text_channel_constructed (GObject *obj)
 {
-  GObject *obj;
   TpsipTextChannelPrivate *priv;
+  TpBaseConnection *base_conn;
+  TpHandleRepoIface *contact_handles;
   DBusGConnection *bus;
+  GObjectClass *parent_object_class =
+      G_OBJECT_CLASS (tpsip_text_channel_parent_class);
 
-  DEBUG("enter");
+  if (parent_object_class->constructed != NULL)
+    parent_object_class->constructed (obj);
 
-  obj =
-    G_OBJECT_CLASS(tpsip_text_channel_parent_class)->constructor(type,
-							       n_props,
-							       props);
   priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE(TPSIP_TEXT_CHANNEL(obj));
+  base_conn = (TpBaseConnection *) priv->conn;
+  contact_handles = tp_base_connection_get_handles (base_conn,
+      TP_HANDLE_TYPE_CONTACT);
+
+  tp_handle_ref (contact_handles, priv->handle);
+
+  g_assert (priv->initiator != 0);
+  tp_handle_ref (contact_handles, priv->initiator);
 
   tpsip_connection_connect_auth_handler (priv->conn, TPSIP_EVENT_TARGET (obj));
 
@@ -173,8 +205,6 @@ GObject *tpsip_text_channel_constructor(GType type,
 
   bus = tp_get_bus();
   dbus_g_connection_register_g_object(bus, priv->object_path, obj);
-
-  return obj;
 }
 
 
@@ -192,6 +222,26 @@ static void tpsip_text_channel_finalize(GObject *object);
 static void
 tpsip_text_channel_class_init(TpsipTextChannelClass *klass)
 {
+  static TpDBusPropertiesMixinPropImpl channel_props[] = {
+      { "ChannelType", "channel-type", NULL },
+      { "Interfaces", "interfaces", NULL },
+      { "TargetHandleType", "handle-type", NULL },
+      { "TargetHandle", "handle", NULL },
+      { "TargetID", "target-id", NULL },
+      { "InitiatorHandle", "initiator-handle", NULL },
+      { "InitiatorID", "initiator-id", NULL },
+      { "Requested", "requested", NULL },
+      { NULL }
+  };
+  static TpDBusPropertiesMixinIfaceImpl prop_interfaces[] = {
+      { TP_IFACE_CHANNEL,
+        tp_dbus_properties_mixin_getter_gobject_properties,
+        NULL,
+        channel_props,
+      },
+      { NULL }
+  };
+
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GParamSpec *param_spec;
 
@@ -202,21 +252,11 @@ tpsip_text_channel_class_init(TpsipTextChannelClass *klass)
   object_class->get_property = tpsip_text_channel_get_property;
   object_class->set_property = tpsip_text_channel_set_property;
 
-  object_class->constructor = tpsip_text_channel_constructor;
+  object_class->constructed = tpsip_text_channel_constructed;
 
   object_class->dispose = tpsip_text_channel_dispose;
   object_class->finalize = tpsip_text_channel_finalize;
 
-  param_spec = g_param_spec_object("connection", "TpsipConnection object",
-				   "SIP connection object that owns this "
-				   "SIP media channel object.",
-				   TPSIP_TYPE_CONNECTION,
-				   G_PARAM_CONSTRUCT_ONLY |
-				   G_PARAM_READWRITE |
-				   G_PARAM_STATIC_NICK |
-				   G_PARAM_STATIC_BLURB);
-  g_object_class_install_property(object_class, PROP_CONNECTION, param_spec);
-
   g_object_class_override_property (object_class, PROP_HANDLE_TYPE,
       "handle-type");
   g_object_class_override_property (object_class, PROP_HANDLE, "handle");
@@ -224,18 +264,67 @@ tpsip_text_channel_class_init(TpsipTextChannelClass *klass)
       "object-path");
   g_object_class_override_property (object_class, PROP_CHANNEL_TYPE,
       "channel-type");
+
+  g_object_class_override_property (object_class, PROP_CHANNEL_DESTROYED,
+      "channel-destroyed");
+  g_object_class_override_property (object_class, PROP_CHANNEL_PROPERTIES,
+      "channel-properties");
+
+  param_spec = g_param_spec_object("connection", "TpsipConnection object",
+      "SIP connection object that owns this SIP media channel object.",
+      TPSIP_TYPE_CONNECTION,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property(object_class, PROP_CONNECTION, param_spec);
+
+  param_spec = g_param_spec_boxed ("interfaces", "Extra D-Bus interfaces",
+      "Addition Channel.Interface.* interfaces", G_TYPE_STRV,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_INTERFACES, param_spec);
+
+  param_spec = g_param_spec_string ("target-id", "Peer's SIP URI",
+      "The URI string obtained by inspecting the peer handle",
+      NULL,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_TARGET_ID, param_spec);
+
+  param_spec = g_param_spec_uint ("initiator-handle", "Initiator's handle",
+      "The contact who initiated the channel",
+      0, G_MAXUINT32, 0,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_INITIATOR_HANDLE,
+      param_spec);
+
+  param_spec = g_param_spec_string ("initiator-id", "Initiator's URI",
+      "The string obtained by inspecting the initiator-handle",
+      NULL,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_INITIATOR_ID,
+      param_spec);
+
+  param_spec = g_param_spec_boolean ("requested", "Requested?",
+      "True if this channel was requested by the local user",
+      FALSE,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_REQUESTED, param_spec);
+
+  klass->dbus_props_class.interfaces =
+      prop_interfaces;
+  tp_dbus_properties_mixin_class_init (object_class,
+      G_STRUCT_OFFSET (TpsipTextChannelClass, dbus_props_class));
 }
 
 static void
 tpsip_text_channel_get_property(GObject *object,
-				   guint property_id,
-				   GValue *value,
-				   GParamSpec *pspec)
+				guint property_id,
+				GValue *value,
+				GParamSpec *pspec)
 {
   TpsipTextChannel *chan = TPSIP_TEXT_CHANNEL(object);
   TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE(chan);
+  TpBaseConnection *base_conn = (TpBaseConnection *) priv->conn;
 
-  switch (property_id) {
+  switch (property_id)
+    {
     case PROP_CONNECTION:
       g_value_set_object(value, priv->conn);
       break;
@@ -248,53 +337,109 @@ tpsip_text_channel_get_property(GObject *object,
       g_value_set_string(value, TP_IFACE_CHANNEL_TYPE_TEXT);
       break;
 
-  case PROP_HANDLE_TYPE:
+    case PROP_HANDLE_TYPE:
       g_value_set_uint(value, TP_HANDLE_TYPE_CONTACT);
-    break;
+      break;
 
-  case PROP_HANDLE:
-    g_value_set_uint(value, priv->handle);
-    break;
+    case PROP_HANDLE:
+      g_value_set_uint(value, priv->handle);
+      break;
 
-  default:
-    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
-    break;
-  }
+    case PROP_TARGET_ID:
+      {
+        TpHandleRepoIface *repo = tp_base_connection_get_handles (
+            base_conn, TP_HANDLE_TYPE_CONTACT);
+
+        g_value_set_string (value, tp_handle_inspect (repo, priv->handle));
+      }
+      break;
+
+    case PROP_INITIATOR_HANDLE:
+      g_value_set_uint (value, priv->initiator);
+      break;
+
+    case PROP_INITIATOR_ID:
+      {
+        TpHandleRepoIface *repo = tp_base_connection_get_handles (
+            base_conn, TP_HANDLE_TYPE_CONTACT);
+
+        g_assert (priv->initiator != 0);
+        g_value_set_string (value,
+            tp_handle_inspect (repo, priv->initiator));
+      }
+      break;
+
+    case PROP_REQUESTED:
+      g_value_set_boolean (value, (priv->initiator == base_conn->self_handle));
+      break;
+
+    case PROP_CHANNEL_DESTROYED:
+      g_value_set_boolean (value, priv->closed);
+      break;
+
+    case PROP_CHANNEL_PROPERTIES:
+      g_value_take_boxed (value,
+          tp_dbus_properties_mixin_make_properties_hash (object,
+              TP_IFACE_CHANNEL, "ChannelType",
+              TP_IFACE_CHANNEL, "TargetHandleType",
+              TP_IFACE_CHANNEL, "TargetHandle",
+              TP_IFACE_CHANNEL, "TargetID",
+              TP_IFACE_CHANNEL, "InitiatorHandle",
+              TP_IFACE_CHANNEL, "InitiatorID",
+              TP_IFACE_CHANNEL, "Requested",
+              TP_IFACE_CHANNEL, "Interfaces",
+              NULL));
+      break;
+
+    case PROP_INTERFACES:
+      g_value_set_static_boxed (value, tpsip_text_channel_interfaces);
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+    }
 }
 
 static void
 tpsip_text_channel_set_property(GObject *object,
-			      guint property_id,
-			      const GValue *value,
-			      GParamSpec *pspec)
+			        guint property_id,
+			        const GValue *value,
+			        GParamSpec *pspec)
 {
   TpsipTextChannel *chan = TPSIP_TEXT_CHANNEL (object);
   TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (chan);
 
-  switch (property_id) {
-  case PROP_CONNECTION:
-    priv->conn = g_value_get_object (value);
-    break;
-    
-  case PROP_OBJECT_PATH:
-    g_assert (priv->object_path == NULL);
-    priv->object_path = g_value_dup_string (value);
-    break;
+  switch (property_id)
+    {
+    case PROP_CONNECTION:
+      priv->conn = g_value_get_object (value);
+      break;
 
-  case PROP_CHANNEL_TYPE:
-  case PROP_HANDLE_TYPE:
-    /* this property is writable in the interface, but not actually
-     * meaningfully changable on this channel, so we do nothing */
-    break;
+    case PROP_OBJECT_PATH:
+      g_assert (priv->object_path == NULL);
+      priv->object_path = g_value_dup_string (value);
+      break;
 
-  case PROP_HANDLE:
-    priv->handle = g_value_get_uint(value);
-    break;
-    
-  default:
-    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
-    break;
-  }
+    case PROP_CHANNEL_TYPE:
+    case PROP_HANDLE_TYPE:
+      /* this property is writable in the interface, but not actually
+       * meaningfully changable on this channel, so we do nothing */
+      break;
+
+    case PROP_HANDLE:
+      /* we don't ref it here because we don't necessarily have access to the
+       * contact repo yet - instead we ref it in the constructed */
+      priv->handle = g_value_get_uint(value);
+      break;
+
+    case PROP_INITIATOR_HANDLE:
+      /* similarly we can't ref this yet */
+      priv->initiator = g_value_get_uint (value);
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+    }
 }
 
 static void
@@ -302,6 +447,7 @@ tpsip_text_channel_dispose(GObject *object)
 {
   TpsipTextChannel *self = TPSIP_TEXT_CHANNEL (object);
   TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_handles;
 
   if (priv->dispose_has_run)
     return;
@@ -309,39 +455,54 @@ tpsip_text_channel_dispose(GObject *object)
   priv->dispose_has_run = TRUE;
 
   if (!priv->closed)
-    tpsip_text_channel_close (self);
+    {
+      priv->closed = TRUE;
+      tp_svc_channel_emit_closed (self);
+    }
 
-  /* release any references held by the object here */
+  contact_handles = tp_base_connection_get_handles (
+      (TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
+
+  tp_handle_unref (contact_handles, priv->handle);
+
+  if (priv->initiator != 0)
+    tp_handle_unref (contact_handles, priv->initiator);
 
   if (G_OBJECT_CLASS (tpsip_text_channel_parent_class)->dispose)
     G_OBJECT_CLASS (tpsip_text_channel_parent_class)->dispose (object);
 }
 
+static void
+zap_pending_messages (GQueue *pending_messages,
+                      TpHandleRepoIface *contact_handles)
+{
+  g_queue_foreach (pending_messages,
+      (GFunc) tpsip_text_pending_free, contact_handles);
+  g_queue_clear (pending_messages);
+}
+
 static void
 tpsip_text_channel_finalize(GObject *object)
 {
   TpsipTextChannel *self = TPSIP_TEXT_CHANNEL (object);
   TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_handles;
 
-  DEBUG("enter");
+  contact_handles = tp_base_connection_get_handles (
+      (TpBaseConnection *)priv->conn, TP_HANDLE_TYPE_CONTACT);
 
-  if (!g_queue_is_empty (priv->pending_messages))
-    {
-      g_warning ("zapping %u pending incoming messages",
-                 g_queue_get_length (priv->pending_messages));
-      g_queue_foreach (priv->pending_messages,
-                       _tpsip_text_pending_free_walk, NULL);
-    }
+  /* XXX: could have responded to the requests with e.g. 480,
+   * but generating network traffic upon abnormal channel termination
+   * does not sound like a good idea */
+  DEBUG ("%u pending incoming messages",
+      g_queue_get_length (priv->pending_messages));
+  zap_pending_messages (priv->pending_messages, contact_handles);
   g_queue_free (priv->pending_messages);
 
-  if (!g_queue_is_empty (priv->messages_to_be_acknowledged))
-    {
-      g_message ("zapping %u pending outgoing message requests",
-                 g_queue_get_length (priv->messages_to_be_acknowledged));
-      g_queue_foreach (priv->messages_to_be_acknowledged,
-                       _tpsip_text_pending_free_walk, NULL);
-    }
-  g_queue_free (priv->messages_to_be_acknowledged);
+  DEBUG ("%u pending outgoing message requests",
+      g_queue_get_length (priv->sending_messages));
+  zap_pending_messages (priv->sending_messages, contact_handles);
+  g_queue_free (priv->sending_messages);
 
   g_free (priv->object_path);
 
@@ -351,7 +512,7 @@ tpsip_text_channel_finalize(GObject *object)
 static gint tpsip_pending_message_compare(gconstpointer msg, gconstpointer id)
 {
   TpsipTextPendingMessage *message = (TpsipTextPendingMessage *)(msg);
-  return (message->id != GPOINTER_TO_INT(id));
+  return (message->id != GPOINTER_TO_UINT(id));
 }
 
 static gint tpsip_acknowledged_messages_compare(gconstpointer msg,
@@ -370,19 +531,16 @@ static gint tpsip_acknowledged_messages_compare(gconstpointer msg,
  */
 static void
 tpsip_text_channel_acknowledge_pending_messages(TpSvcChannelTypeText *iface,
-					      const GArray *ids,
-					      DBusGMethodInvocation *context)
+                                                const GArray *ids,
+                                                DBusGMethodInvocation *context)
 {
-  TpsipTextChannel *obj = TPSIP_TEXT_CHANNEL (iface);
-  TpsipTextChannelPrivate *priv;
+  TpsipTextChannel *chan = TPSIP_TEXT_CHANNEL (iface);
+  TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE(chan);
   TpHandleRepoIface *contact_repo;
   GList **nodes;
   TpsipTextPendingMessage *msg;
   guint i;
 
-  DEBUG("enter");
-  
-  priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE(obj);
   contact_repo = tp_base_connection_get_handles (
       (TpBaseConnection *)(priv->conn), TP_HANDLE_TYPE_CONTACT);
 
@@ -393,7 +551,7 @@ tpsip_text_channel_acknowledge_pending_messages(TpSvcChannelTypeText *iface,
       guint id = g_array_index(ids, guint, i);
 
       nodes[i] = g_queue_find_custom (priv->pending_messages,
-                                      GINT_TO_POINTER (id),
+                                      GUINT_TO_POINTER (id),
                                       tpsip_pending_message_compare);
 
       if (nodes[i] == NULL)
@@ -420,8 +578,9 @@ tpsip_text_channel_acknowledge_pending_messages(TpSvcChannelTypeText *iface,
 
       g_queue_remove (priv->pending_messages, msg);
 
-      tp_handle_unref (contact_repo, msg->sender);
-      _tpsip_text_pending_free (msg);
+      tpsip_text_pending_respond (msg);
+
+      tpsip_text_pending_free (msg, contact_repo);
     }
 
   g_free(nodes);
@@ -431,32 +590,84 @@ tpsip_text_channel_acknowledge_pending_messages(TpSvcChannelTypeText *iface,
 
 
 /**
- * tpsip_text_channel_close_async
+ * tpsip_text_channel_close
  *
  * Implements DBus method Close
  * on interface org.freedesktop.Telepathy.Channel
  */
 static void
-tpsip_text_channel_close_async (TpSvcChannel *iface,
-                              DBusGMethodInvocation *context)
+tpsip_text_channel_close (TpSvcChannel *iface,
+                          DBusGMethodInvocation *context)
 {
-  tpsip_text_channel_close (TPSIP_TEXT_CHANNEL(iface));
+  TpsipTextChannel *self = TPSIP_TEXT_CHANNEL (iface);
+  TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE(self);
+
+  if (priv->closed)
+    {
+      DEBUG ("already closed, doing nothing");
+    }
+  else
+    {
+      if (g_queue_is_empty (priv->pending_messages))
+        {
+          DEBUG ("actually closing, no pending messages");
+          priv->closed = TRUE;
+        }
+      else
+        {
+          GList *cur;
+
+          DEBUG ("not really closing, there are pending messages left");
+
+          for (cur = g_queue_peek_head_link (priv->pending_messages);
+               cur != NULL;
+               cur = cur->next)
+            {
+              TpsipTextPendingMessage *msg = cur->data;
+              msg->flags |= TP_CHANNEL_TEXT_MESSAGE_FLAG_RESCUED;
+            }
+
+          if (priv->initiator != priv->handle)
+            {
+              TpHandleRepoIface *contact_repo = tp_base_connection_get_handles
+                  ((TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
+
+              g_assert (priv->initiator != 0);
+              g_assert (priv->handle != 0);
+
+              tp_handle_unref (contact_repo, priv->initiator);
+              priv->initiator = priv->handle;
+              tp_handle_ref (contact_repo, priv->initiator);
+            }
+        }
+      tp_svc_channel_emit_closed (self);
+    }
   tp_svc_channel_return_from_close (context);
 }
 
-void
-tpsip_text_channel_close (TpsipTextChannel *self)
+/**
+ * tpsip_text_channel_destroy
+ *
+ * Implements D-Bus method Destroy
+ * on interface org.freedesktop.Telepathy.Channel.Interface.Destroyable
+ */
+static void
+tpsip_text_channel_destroy (TpSvcChannelInterfaceDestroyable *iface,
+                            DBusGMethodInvocation *context)
 {
-  TpsipTextChannelPrivate *priv;
+  TpsipTextChannel *self = TPSIP_TEXT_CHANNEL (iface);
+  TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_handles;
 
-  DEBUG("enter");
+  contact_handles = tp_base_connection_get_handles (
+      (TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
 
-  priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE(self);
-  if (!priv->closed)
-    {
-      priv->closed = TRUE;
-      tp_svc_channel_emit_closed (self);
-    }
+  /* Make sure there are no pending messages for Close to get excited about */
+  zap_pending_messages (priv->pending_messages, contact_handles);
+
+  /* Close() and Destroy() have the same signature, so we can safely
+   * chain to the other function now */
+  tpsip_text_channel_close ((TpSvcChannel *) self, context);
 }
 
 /**
@@ -508,10 +719,9 @@ static void
 tpsip_text_channel_get_interfaces(TpSvcChannel *iface,
                                 DBusGMethodInvocation *context)
 {
-  const char *interfaces[] = { NULL };
-
   DEBUG("enter");
-  tp_svc_channel_return_from_get_interfaces (context, interfaces);
+  tp_svc_channel_return_from_get_interfaces (context,
+      tpsip_text_channel_interfaces);
 }
 
 
@@ -549,7 +759,7 @@ tpsip_pending_message_list_add (GPtrArray *list, TpsipTextPendingMessage *msg)
                           1, msg->timestamp,
                           2, msg->sender,
                           3, msg->type,
-                          4, 0 /* msg->flags */,
+                          4, msg->flags,
                           5, msg->text,
                           G_MAXUINT);
 
@@ -581,9 +791,15 @@ tpsip_text_channel_list_pending_messages(TpSvcChannelTypeText *iface,
 
   if (clear)
     {
+      TpHandleRepoIface *contact_repo = tp_base_connection_get_handles (
+          (TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
       while ((cur = g_queue_pop_head_link (priv->pending_messages)) != NULL)
-        tpsip_pending_message_list_add (messages,
-                                        (TpsipTextPendingMessage *) cur->data);
+        {
+          TpsipTextPendingMessage * msg = (TpsipTextPendingMessage *) cur->data;
+          tpsip_pending_message_list_add (messages, msg);
+          tpsip_text_pending_respond (msg);
+          tpsip_text_pending_free (msg, contact_repo);
+        }
     }
   else
     {
@@ -652,13 +868,13 @@ tpsip_text_channel_send(TpSvcChannelTypeText *iface,
 	      SIPTAG_PAYLOAD_STR(text),
 	      TAG_END());
 
-  msg = _tpsip_text_pending_new();
+  msg = _tpsip_text_pending_new0 ();
   msg->nh = msg_nh;
   msg->text = g_strdup(text);
   msg->type = type;
   msg->timestamp = time(NULL);
 
-  g_queue_push_tail (priv->messages_to_be_acknowledged, msg);
+  g_queue_push_tail (priv->sending_messages, msg);
 
   DEBUG("message queued for delivery with timestamp %u", (guint)msg->timestamp);
 
@@ -673,12 +889,15 @@ tpsip_text_channel_nua_r_message_cb (TpsipTextChannel *self,
 {
   TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (self);
   TpsipTextPendingMessage *msg;
+  TpHandleRepoIface *contact_repo;
   TpChannelTextSendError send_error;
   GList *node;
 
-  DEBUG("enter");
+  /* ignore provisional responses */
+  if (ev->status < 200)
+    return TRUE;
 
-  node = g_queue_find_custom (priv->messages_to_be_acknowledged,
+  node = g_queue_find_custom (priv->sending_messages,
                               ev->nua_handle,
 			      tpsip_acknowledged_messages_compare);
 
@@ -738,44 +957,52 @@ tpsip_text_channel_nua_r_message_cb (TpsipTextChannel *self,
       }
 
     tp_svc_channel_type_text_emit_send_error (self,
-	send_error, msg->timestamp, msg->type, msg->text);  
+	send_error, msg->timestamp, msg->type, msg->text);
   }
 
-  g_queue_remove(priv->messages_to_be_acknowledged, msg);
-  _tpsip_text_pending_free(msg);
+  g_queue_remove(priv->sending_messages, msg);
+
+  contact_repo = tp_base_connection_get_handles (
+      (TpBaseConnection *)(priv->conn), TP_HANDLE_TYPE_CONTACT);
+
+  tpsip_text_pending_free(msg, contact_repo);
 
   return TRUE;
 }
 
 void tpsip_text_channel_receive(TpsipTextChannel *chan,
-			      TpHandle sender,
-			      const char *message)
+                                nua_t *nua,
+                                nua_handle_t *nh,
+                                TpHandle sender,
+                                const char *text,
+                                gsize len)
 {
+  TpsipTextChannelPrivate *priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (chan);
   TpsipTextPendingMessage *msg;
-  TpsipTextChannelPrivate *priv;
   TpHandleRepoIface *contact_repo;
 
-  DEBUG("enter");
-
-  g_assert(chan != NULL);
-  g_assert(TPSIP_IS_TEXT_CHANNEL(chan));
-  priv = TPSIP_TEXT_CHANNEL_GET_PRIVATE (chan);
-  contact_repo = tp_base_connection_get_handles (
-      (TpBaseConnection *)(priv->conn), TP_HANDLE_TYPE_CONTACT);
-
-  msg = _tpsip_text_pending_new();
+  msg = _tpsip_text_pending_new0 ();
 
   msg->id = priv->recv_id++;
   msg->timestamp = time(NULL);
+  msg->nh = nh;
   msg->sender = sender;
   msg->type = TP_CHANNEL_TEXT_MESSAGE_TYPE_NORMAL;
-  msg->text = g_strdup(message);
+  msg->text = g_strndup (text, len);
+
+  nua_save_event (nua, msg->saved_event);
 
   g_queue_push_tail(priv->pending_messages, msg);
 
-  tp_handle_ref (contact_repo, msg->sender);
+  nua_handle_ref (nh);
 
-  DEBUG("received message: %s", message);
+  contact_repo = tp_base_connection_get_handles (
+      (TpBaseConnection *)(priv->conn), TP_HANDLE_TYPE_CONTACT);
+
+  tp_handle_ref (contact_repo, sender);
+
+  DEBUG("received message id %u, now %u pending",
+        msg->id, g_queue_get_length (priv->pending_messages));
 
   tp_svc_channel_type_text_emit_received ((TpSvcChannelTypeText *)chan,
       msg->id, msg->timestamp, msg->sender, msg->type,
@@ -783,8 +1010,15 @@ void tpsip_text_channel_receive(TpsipTextChannel *chan,
 }
 
 static void
-event_target_iface_init (gpointer g_iface, gpointer iface_data)
+destroyable_iface_init (gpointer g_iface,
+                        gpointer iface_data)
 {
+  TpSvcChannelInterfaceDestroyableClass *klass = g_iface;
+
+#define IMPLEMENT(x) tp_svc_channel_interface_destroyable_implement_##x (\
+    klass, tpsip_text_channel_##x)
+  IMPLEMENT(destroy);
+#undef IMPLEMENT
 }
 
 static void
@@ -792,12 +1026,12 @@ channel_iface_init(gpointer g_iface, gpointer iface_data)
 {
   TpSvcChannelClass *klass = (TpSvcChannelClass *)g_iface;
 
-#define IMPLEMENT(x, suffix) tp_svc_channel_implement_##x (\
-    klass, tpsip_text_channel_##x##suffix)
-  IMPLEMENT(close,_async);
-  IMPLEMENT(get_channel_type,);
-  IMPLEMENT(get_handle,);
-  IMPLEMENT(get_interfaces,);
+#define IMPLEMENT(x) tp_svc_channel_implement_##x (\
+      klass, tpsip_text_channel_##x)
+  IMPLEMENT(close);
+  IMPLEMENT(get_channel_type);
+  IMPLEMENT(get_handle);
+  IMPLEMENT(get_interfaces);
 #undef IMPLEMENT
 }
 
